"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

Child behaviour helpers.

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>.
"""
from pyzora.enums import *
from typing import Literal
from itertools import starmap


_NORM_METHODS = (str.upper, lambda x: x)


def _normalise_name(region: GameRegion, name: str) -> str:
    """Return a normalised version of the child's name.
    If the region is not JP, then this returns the unchanged string.

    :param region: The region to use for this operation.
    :type region: GameRegion
    :param name: The child's name.
    :type name: str
    :return: The normalised name.
    :rtype: str
    :meta private:"""
    return _NORM_METHODS[region](name)


def _get_value_for_child_name(region: GameRegion, name: str):
    if len(name) > 5:
        raise ValueError(f"invalid child name : must be at most 5 characters long (got a {len(name)}-character string instead)")
    b_name = bytes(_normalise_name(region, name), "utf-8")
    value = sum(map((0xF).__and__, b_name)) - 3
    while value > 255:
        value -= 3
    return value + 4


_ChildBehaviourArgs = tuple[RupeesGiven, SleepMethod, ChildQuestion, ChildKind] | tuple[RupeesGiven, SleepMethod]
_CHILDBEVHAVIOURTYPES=(RupeesGiven, SleepMethod, ChildQuestion, ChildKind)


def get_child_behaviour_value(
        region: GameRegion,
        name: str,
        *args: _ChildBehaviourArgs
) -> int:
    """Get the raw integer value for the child's behaviour based on the answer to the questions.

    :param region: The region to use for this (puts the name to uppercase if the region is Japan).
    :type region: GameRegion
    :param name: The child's name.
    :type name: str
    :param args: Contains in this order, how many rupees you gave to Blossom, the sleep method you suggested her to use, the answer you gave to the kid's question and what kind of child you were.
    :type args: RupeesGiven, SleepMethod, ChildQuestion, ChildKind
    :return: The raw behaviour value (does not go above 255).
    :rtype: int"""
    if not len(args):
        return _get_value_for_child_name(region, name)
    if len(args) not in {2, 4}:
        raise TypeError(f"unexpected arguments (starting at the {('third', 'fifth')[len(args) > 4]} value) : {args}")
    if not all(starmap(isinstance, zip(args, _CHILDBEVHAVIOURTYPES))):
        raise TypeError(f"wrong argument order or unexpected arguments : {tuple(map(lambda obj: type(obj).__qualname__, args))}")
    return _get_value_for_child_name(region, name) + sum(args)


_Byte = Literal[*range(256)]  # writing all values from 0 to 256 would be slow and ugly


def get_behaviour(value: _Byte) -> ChildBehaviour:
    """Return the behaviour type for a raw integer value.

    :param value: The raw behaviour value.
    :type value: int
    :raise ValueError: if the given value is not between 0 and 255.
    :return: The corresponding child behaviour.
    :rtype: ChildBehaviour"""
    if value > 255 or value < 0:
        raise ValueError(f"integer value is out of bounds : {value}")
    if value == 0:
        return ChildBehaviour.NONE
    if value < 6:
        return ChildBehaviour.CURIOUS
    if value < 11:
        return ChildBehaviour.SHY
    return ChildBehaviour.BOUNCY



"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

Exceptions raised by library functions.

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>.
"""


class SecretError(Exception):
    """Something is wrong with the secret."""


class ChecksumError(SecretError):
    """Wrong secret checksum."""


class NotAGameCodeError(SecretError):
    """The given secret is not a game code."""


class NotARingCodeError(SecretError):
    """The given secret is not a ring password."""


class NotAMemoryCodeError(SecretError):
    """The given secret is not a memory password."""



"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

Ring types used by the library. They are available for any checks you might want
to do on the secrets.

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>.
"""
from more_itertools import first


class __NoRingsType:
    """A singleton representing a game with no rings obtained."""
    __INST = 0
    __slots__ = ()

    def __new__(cls):
        if not cls.__INST:
            cls.__INST += 1
            return super().__new__(cls)
        return NoRings

    def __del__(self):
        self.__class__.__INST -= 1

    def __or__(self, other):
        return int(self) | other

    def __ror__(self, other):
        return self.__or__(other)

    def __int__(self):
        return 0

    def __bool__(self):
        return False


NoRings = __NoRingsType()


class RingType:
    """Represents one of the rings scattered in Labrynna and Holodrum.

    RingType instances DO NOT store the modifiers used by the rings when
    equipped nor their randomisation tiers. You can check this info """
    def __init__(self, integer, name, description):
        self.integer, self.name, self.description = integer, name, description

    def __repr__(self):
        return self.name

    def __int__(self):
        return self.integer

    def __and__(self, other):
        return int(self) & other

    def __rand__(self, other):
        return self.__and__(other)

    def __or__(self, other):
        return int(self) | other

    def __ror__(self, other):
        return self.__or__(other)

    def __eq__(self, other):
        if type(self) is type(other):
            return self is other
        return int(self) == int(other)

    @classmethod
    def from_integer(cls, integer: int):
        """Returns the first ring type with the corresponding integer.
        If nothing matches, NoRings is returned instead."""
        if integer > int(AllRings):
            raise ValueError("given value exceeds AllRings' value")
        elif integer < 0:
            raise ValueError("given value is negative")
        return first(filter(lambda tp: integer == tp.integer, RING_TYPES), NoRings)


class __AllRingsType:
    """A singleton representing all obtainable rings in a game."""
    __INST = 0
    __slots__ = ()

    def __new__(cls):
        if not cls.__INST:
            cls.__INST += 1
            return super().__new__(cls)
        return AllRings

    def __del__(self):
        self.__class__.__INST -= 1

    def __or__(self, other):
        return int(self) | other

    __ror__ = __or__

    def __int__(self):
        return 18446744073709551615

    def __bool__(self):
        return True


AllRings = __AllRingsType()

FRIENDSHIP = RingType(0x1, "Friendship Ring", "Symbol of a meeting")
POWER_1 = RingType(0x2, "Power Ring L-1", "Sword damage ▲\nDamage taken ▲")
POWER_2 = RingType(0x4, "Power Ring L-2", "Sword damage ▲▲\nDamage taken ▲▲")
POWER_3 = RingType(0x8, "Power Ring L-3", "Sword damage ▲▲▲\nDamage taken ▲▲▲")
ARMOUR_1 = RingType(0x10, "Armour Ring L-1", "Sword Damage ▼\nDamage taken ▼")
ARMOUR_2 = RingType(0x20, "Armour Ring L-2", "Sword Damage ▼▼\nDamage taken ▼▼")
ARMOUR_3 = RingType(0x40, "Armour Ring L-3", "Sword Damage ▼▼▼\nDamage taken ▼▼▼")
RED = RingType(0x80, "Red Ring", "Sword Damage x2")
BLUE = RingType(0x100, "Blue Ring", "Damage taken reduced by 1/2")
GREEN = RingType(0x200, "Green Ring", "Damage taken down by 25%\nSword damage up by 50%")
CURSED = RingType(0x400, "Cursed Ring", "1/2 sword damage\nx2 damage taken")
EXPERTS_RING = RingType(0x800, "Expert's Ring", "Punch when unequipped")
BLAST_RING = RingType(0x1000, "Blast Ring", "Bomb damage ▲")
RANG1 = RingType(0x2000, "Rang Ring L-1", "Boomerang damage ▲")
GBA_TIME = RingType(0x4000, "GBA Time Ring", "Life Advanced!")
MAPLES_RING = RingType(0x8000, "Maple's Ring", "Maple meetings ▲")
STEADFAST = RingType(0x10000, "Steadfast Ring", "Get knocked back less")
PEGASUS = RingType(0x20000, "Pegasus Ring", "Lengthen Pegasus Seed effect")
TOSS_RING = RingType(0x40000, "Toss Ring", "Throwing distance ▲")
HEART1 = RingType(0x80000, "Heart Ring L-1", "Slowly recover lost Hearts")
HEART2 = RingType(0x100000, "Heart Ring L-2", "Recover lost Hearts")
SWIMMERS_RING = RingType(0x200000, "Swimmer's Ring", "Swimming speed ▲")
CHARGE_RING = RingType(0x400000, "Charge Ring", "Spin Attack charges quickly")
LIGHT1 = RingType(0x800000, "Light Ring L-1", "Sword beams at -2 Hearts")
LIGHT2 = RingType(0x1000000, "Light Ring L-2", "Sword beams at -3 Hearts")
BOMBERS_RING = RingType(0x2000000, "Bomber's Ring", "Set two Bombs at once")
GREEN_LUCK = RingType(0x4000000, "Green Luck Ring", "1/2 damage from traps")
BLUE_LUCK = RingType(0x8000000, "Blue Luck Ring", "1/2 damage from beams")
GOLD_LUCK = RingType(0x10000000, "Gold Luck Ring", "1/2 damage from falls")
RED_LUCK = RingType(0x20000000, "Red Luck Ring", "1/2 damage from spiked floors")
GREEN_HOLY = RingType(0x40000000, "Green Holy Ring", "No damage from electricity")
BLUE_HOLY = RingType(0x80000000, "Blue Holy Ring", "No damage from Zora's fire")
RED_HOLY = RingType(0x100000000, "Red Holy Ring", "No damage from small rocks")
SNOWSHOE = RingType(0x200000000, "Snowshoe Ring", "No sliding on ice")
ROCS_RING = RingType(0x400000000, "Roc's Ring", "Cracked floors don't crumble")
QUICKSAND = RingType(0x800000000, "Quicksand Ring", "No sinking in quicksand")
RED_JOY = RingType(0x1000000000, "Red Joy Ring", "Beasts drop double Rupees")
BLUE_JOY = RingType(0x2000000000, "Blue Joy Ring", "Beasts drop double Hearts")
GOLD_JOY = RingType(0x4000000000, "Gold Joy Ring", "Find double items")
GREEN_JOY = RingType(0x8000000000, "Green Joy Ring", "Find double Ore Chunks")
DISCOVERY = RingType(0x10000000000, "Discovery Ring", "Sense soft earth nearby")
RANG2 = RingType(0x20000000000, "Rang Ring L-2", "Boomerang damage ▲▲")
OCTO = RingType(0x40000000000, "Octo Ring", "Become an Octorok")
MOBLIN = RingType(0x80000000000, "Moblin Ring", "Become a Moblin")
LIKE_LIKE = RingType(0x100000000000, "Like Like Ring", "Become a Like-Like")
SUBROSIAN = RingType(0x200000000000, "Subrosian Ring", "Become a Subrosian")
GEN1 = RingType(0x400000000000, "First Gen Ring", "Become something")
SPIN = RingType(0x800000000000, "Spin Ring", "Double Spin Attack")
BOMBPROOF = RingType(0x1000000000000, "Bombproof Ring", "No damage from your own Bombs")
ENERGY = RingType(0x2000000000000, "Energy Ring", "Beam replaces Spin Attack")
DOUBLE_EDGED = RingType(0x4000000000000, "Dbl. Edge Ring", "Sword damage ▲ but you get hurt")
GBA_NATURE = RingType(0x8000000000000, "GBA Nature Ring", "Life Advanced!")
SLAYER = RingType(0x10000000000000, "Slayer's Ring", "1000 beasts slain")
RUPEE = RingType(0x20000000000000, "Rupee Ring", "10,000 Rupees collected")
VICTORY = RingType(0x40000000000000, "Victory Ring", "The Evil King Ganon defeated")
SIGN = RingType(0x80000000000000, "Sign Ring", "100 signs broken")
HUNDREDTH = RingType(0x100000000000000, "100th Ring", "100 rings appraised")
WHISP = RingType(0x200000000000000, "Whisp Ring", "No effect from jinxes")
GASHA = RingType(0x400000000000000, "Gasha Ring", "Grow great Gasha Trees")
PEACE = RingType(0x800000000000000, "Peace Ring", "No explosion if holding Bomb")
ZORA = RingType(0x1000000000000000, "Zora Ring", "Dive without breathing")
FIST = RingType(0x2000000000000000, "Fist Ring", "Punch when not equipped")
WHIMSICAL = RingType(0x4000000000000000, "Whimsical Ring", "Sword damage ▼ Sometimes deadly")
PROTECTION = RingType(0x8000000000000000, "Protection Ring", "Damage taken is always one Heart")


RING_TYPES = (FRIENDSHIP, POWER_1, POWER_2, POWER_3, ARMOUR_1,
              ARMOUR_2, ARMOUR_3, RED, BLUE, GREEN, CURSED, EXPERTS_RING,
              BLAST_RING, RANG1, GBA_TIME, MAPLES_RING, STEADFAST,
              PEGASUS, TOSS_RING, HEART1, HEART2, SWIMMERS_RING,
              CHARGE_RING, LIGHT1, LIGHT2, BOMBERS_RING, GREEN_LUCK,
              BLUE_LUCK, GOLD_LUCK, RED_LUCK, GREEN_HOLY, BLUE_HOLY,
              RED_HOLY, SNOWSHOE, ROCS_RING, QUICKSAND, RED_JOY,
              BLUE_JOY, GOLD_JOY, GREEN_JOY, DISCOVERY, RANG2, OCTO,
              MOBLIN, LIKE_LIKE, SUBROSIAN, GEN1, SPIN, BOMBPROOF,
              ENERGY, DOUBLE_EDGED, GBA_NATURE, SLAYER, RUPEE,
              VICTORY, SIGN, HUNDREDTH, WHISP, GASHA, PEACE, ZORA, FIST,
              WHIMSICAL, PROTECTION)



"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

All enums used by the library. Check each one's documentation for more details.

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>."""
from enum import IntEnum


class GameRegion(IntEnum):
    """A secret's game region.
    Note that US/PAL secrets do NOT work with Japanese ROMS,
    and vice-versa."""

    JP = 0
    """Indicates the current operation should be realised with the Japanese ciphers."""

    US_PAL = 1
    """Indicates the current operation should be achieved using the international ciphers (
    US and PAL cartridges use the same characters and encryption-decryption methods)."""


class TargetGame(IntEnum):
    """The game targeted by the secret, represented by one bit in the secret
    (position differs depending on the secret type).

    Secrets meant for one game will not work with the other one.
    This is only relevant for game and memory secrets, since ring secrets are
    game-independent."""

    AGES = 0
    """Indicates the current secret's target game is Oracle of Ages."""

    SEASONS = 1
    """Indicates the current secret's target game is Oracle of Seasons."""


class ObtainedCompanion(IntEnum):
    """The animal friend tied to this secret.

    The NONE value technically is possible to encounter in Hero's Secrets,
    but its use is otherwise discouraged, bar non-linked secrets.
    You might end up with unexpected results otherwise."""

    NONE = 0
    """The player has no animal friend yet. This value normally only gets used in
    Hero's Secrets (non-Hero secrets should also work with this value, provided
    they aren't linked)."""

    RICKY = 0x0b
    """The player has obtained Ricky the Kangaroo's Flute in the completed file."""

    DIMITRI = 0x0c
    """The player has obtained Dimitri the Dodongo's Flute in the finished save."""

    MOOSH = 0x0d
    """The player obtained Moosh the Flying Bear's Flute in the original file."""


class MemoryEnum(IntEnum):
    """The corresponding memory secret.
    Names are formatted like "A_OR_B", where A is
    the memory in linked Ages and B the memory in Linked Seasons.
    Each member's description explains the name chosen for it."""

    CLOCKSHOP_OR_KINGZORA = 0
    """Ages : in Present Labrynna, just south of the Lynna Shop, an old woman spawns
    and tells you to dig behind the Clock Shop in Horon Village after you complete 
    the Wing Dungeon.
    
    Seasons : a Zora spawns in Hero's Cave after you get the Roc's Feather in 
    Poison Moth's Lair, and tells you to meet King Zora in Labrynna."""

    GRAVEYARD_OR_FAIRY = 1
    """Ages : in Present Labrynna, a Ghini spawns under Syrup's Potion Shop in
    the Yoll Graveyard after you complete Spirit's Grave. It tells you to find another
    Ghini under a flowerless grave in the Western Coast Graveyard in Holodrum.
    
    Seasons : right after you talk with the Maku Tree for the first time, a woman spawns next to Horon Village's fountain. 
    She wants you to tell the "happiness" secret to the Blue Fairy in Fairies' Woods in Labrynna."""

    SUBROSIAN_OR_TROY = 2
    """Ages : after you beat the Crown Dungeon, a Subrosian appears in a cave in eastern
    Rolling Ridge in the present. He asks you to get in touch with his golden brother back in Subrosia.
    
    Seasons : after you complete Poison Moth's Lair, a kid spawns in the Eastern Suburbs.
    He gives you a secret to give to Dr. Troy in Present Labrynna, who went to the Target Carts in Rolling Ridge."""

    DIVER_OR_PLEN = 3
    """Ages : after you beat Mermaid's Cave, a diver spawns above Zora Village in the present. He asks you to tell a 
    secret to the Master Diver, who lives in the Sunken City in Holodrum.
    
    Seasons : once you get the Rod of Seasons infused with the Power of Autumn, a golden Subrosian spawns at Lava 
    Lake. He wants you to go thank Mayor Plen in Lynna City in his stead."""

    SMITH_OR_LIBRARY = 4
    """Ages : after you complete the Crown Dungeon, a Subrosian spawns next to the
    Mysterious Tree at eastern Rolling Ridge in the past. He gives you a secret to 
    tell to the Subrosian Smithy.
    
    Seasons : after you help the Piratians, a Ghini spawns inside the deserted house
    on the Western Coast. It tells you a secret to carry to the Eyeglass Isle Library in past Labrynna."""

    PIRATE_OR_TOKAY = 5
    """Ages : a little girl spawns just south of the Black Tower in the past after you get the Roc's Feather in the 
    Wing Dungeon. She wants you to give her luck secret to one of the Piratians still stranded on Subrosia, because she
    forgot to gave it to him earlier.
    
    Seasons : after destroying Moblin's Keep and visiting their refuge in the Sunken City, a Moblin will spawn in the fortress's
    ruins and tell you a secret to be passed to the Wild Tokay Museum curator on Crescent Island in the present."""

    TEMPLE_OR_MAMAMU = 6
    """Ages : after you complete the Wing Dungeon, a Fairy appears over its ruins in Fairies' Woods. She has a secret for you
    to give to another Fairy living in the innermost room of the Temple of Seasons.
    
    Seasons : once you complete Poison Moth's Lair, an old woman will spawn inside the Floodgate Keeper's house.
    Her secret will have to be carried to Mamamu Yan, who lives in Lynna City in the present."""

    DEKU_OR_TINGLE = 7
    """Ages : after you complete the Spirit's Grave, a Deku Scrub spawns in the Deku Forest in the past. It sings a 
    then-popular song, and the end of it is a secret. You will need to give it to another Deku Scrub living near Moblin's 
    Keep in Holodrum.
    
    Seasons : after completing Snake's Remains, a Fairy will spawn in a cave east of the dungeon. She has a secret
    for you to tell to Tingle, who's living in the Forest of Time in present Labrynna."""

    BIGGORON_OR_ELDER = 8
    """Ages : once you beat the Crown Dungeon, a Goron will spawn in a cave in eastern Rolling Ridge in the present.
    He has a secret you need to pass on to Biggoron, who can be reached atop Goron Mountain in Holodrum.
    
    Seasons : after you cure Biggoron's cold with the Lava Soup, a red Goron will spawn in a cave right under Biggoron.
    He wants you to ask his ancestor (found in the Goron Shooting Gallery) in the past the meaning of a secret he found in 
    his ancestor's diary."""

    RUUL_OR_SYMMETRY = 9
    """Ages : once you beat the Wing Dungeon, Mayor Plen's grandmother appears alongside him in his house. She has a secret 
    you will need to tell to Mayor Ruul in Horon Village.
    
    Seasons : after beating the Snake's Remains, a girl spawns in Holly's house in Holodrum. She wants you to thank the authors
    of her favourite book, the Middle House Twins from Symmetry Village, by telling them a touching secret."""


class ChildBehaviour(IntEnum):
    """Represents Bipin and Blossom's son's state of growth and
    personality based on Link's choices."""
    NONE = 0
    CURIOUS = 1
    SHY = 2
    BOUNCY = 3


class RupeesGiven(IntEnum):
    """Represents how many Rupees you gave to Blossom to help her care for
    her child."""
    ONE = 0
    TEN = 2
    FIFTY = 5
    ONE_HUNDRED_AND_FIFTY = 8


class SleepMethod(IntEnum):
    """Represent the method you recommended to Blossom to help her son sleep."""
    SING = 0
    PLAY = 10


class ChildKind(IntEnum):
    """Represents which personality you told Blossom you had as a kid."""
    NONE = 0
    WEIRD = 1
    QUIET = 5
    HYPERACTIVE = 8


class ChildQuestion(IntEnum):
    """Represents the answer you gave to the child's question.
    The bits set are the same regardless of the question asked."""
    NO_OR_EGG = 0
    YES_OR_CHICKEN = 4



"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

Memory secret class. See MemorySecret's documentation for more details.

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>.
"""
from pyzora.secret import *


class MemorySecret(BaseSecret):
    """A memory secret to transfer between two NPCS in Holodrum and Labrynna."""

    __args = ("game_id", "region", "target_game", "memory", "is_return_secret")
    __game_id = 0
    __target_game = TargetGame.AGES
    __memory = MemoryEnum.CLOCKSHOP_OR_KINGZORA
    __is_return = False

    def __set_memory(self, value: int):
        if value < 0:
            raise ValueError("expected a positive value")
        self.__memory = MemoryEnum(value)

    memory = property(lambda self: self.__memory, __set_memory,
                      doc="""The event tied to this secret. See MemoryEnum for more
                      details.
                      
                      :type: MemoryEnum""")

    def __set_target_game(self, value: TargetGame):
        self.__target_game = TargetGame(value)

    target_game = property(lambda self: self.__target_game, __set_target_game,
                           doc="""The secret's target game. See GameSecret.target_game
                           for more details.
                           
                           :type: TargetGame""")

    def __set_return(self, value: bool):
        if not isinstance(value, (bool, int, float)):
            raise TypeError("expected a boolean value")
        self.__is_return = bool(value)

    is_return_secret = property(lambda self: self.__is_return, __set_return,
                                doc="""Set this if this secret should be returned to the non-completed
                                file.
                                
                                :type: bool""")

    def __init__(self, *args, **kwargs):
        for pos, arg in enumerate(self.__args):
            kwargs.setdefault(arg, None)
            if kwargs.get(arg) is None:
                try:
                    setattr(self, arg, args[pos])
                except LookupError:
                    pass
            else:
                try:
                    setattr(self, arg, kwargs.get(arg))
                except LookupError:
                    pass

    @classmethod
    def load(cls, secret: bytearray | bytes | str, region: GameRegion) -> "MemorySecret":
        """Load a memory secret from bytes or a string.

        :param secret: The secret string/byte array to decode (if secret is an instance of str, then it will be parsed prior to loading).
        :type secret: str or bytearray
        :param region: The region to use when loading the secret.
        :type region: GameRegion
        :raise SecretError: if the given data is not 20 bytes long (only if used with bytes or bytearray, since strings will be parsed and reduced to 20 bytes).
        :raise NotAMemoryCodeError: if the given data is not for a game secret.
        :raise ChecksumError: if the given data's checksum doesn't match the expected one.
        :return: A secret based on the data contained in secret.
        :rtype: MemorySecret"""
        if isinstance(secret, str):
            # Secret string. Parse it before doing anything else.
            # Allows the user to give the region only once
            return cls.load(parse_secret(secret, region), region)
        if len(secret) != 5:
            raise SecretError("secret must contain exactly 5 bytes")
        decoded_bytes = cls.decode_bytes(secret, region)
        decoded_secret = byte_array_to_string(decoded_bytes)
        cloned_bytes = decoded_bytes.copy()
        cloned_bytes[4] = 0
        checksum = calculate_checksum(cloned_bytes)
        if (decoded_bytes[4] & 0xF) != (checksum & 0xF):
            raise ChecksumError(f"the two values ({decoded_bytes[4]} and {checksum}) do not match")
        if decoded_secret[3:5] != "11":
            raise NotAMemoryCodeError("given secret is not a memory code")
        game_id = int(reverse_substring(decoded_secret, 5, 15), 2)
        memory = MemoryEnum(int(reverse_substring(decoded_secret, 20, 4), 2))
        target_game = TargetGame(decoded_secret[24] != decoded_secret[25])
        is_return_secret = bool(int(decoded_secret[24]))
        return MemorySecret(game_id=game_id, region=region, memory=memory,
                            target_game=target_game, is_return_secret=is_return_secret)

    def __bytes__(self):
        if self.target_game:
            cipher = 2 - self.__is_return
        else:
            cipher = 3*self.__is_return
        cipher |= (self.memory & 1) << 2
        cipher = ((self.game_id >> 8) + (self.game_id & 255) + cipher) & 7
        cipher = int(reverse_string(integer_string(cipher).rjust(3, "0")), 2)
        unencoded_secret = "".join((
                integer_string(cipher).rjust(3, "0"), "11",
                reverse_string(integer_string(cipher).rjust(15, "0")),
                reverse_string(integer_string(self.memory).rjust(4, "0"))
        ))
        if self.target_game:
            mask = 1 + self.__is_return
        else:
            mask = 3*self.__is_return
        unencoded_bytes = string_to_byte_array(unencoded_secret)
        unencoded_bytes[4] = calculate_checksum(unencoded_bytes) | (mask << 4)
        return bytes(self._encode_bytes(unencoded_bytes, self.region))



"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>.
"""
from pyzora.enums import *
from pyzora.secret import *
from pyzora.game_secret import *
from pyzora.memory_secret import *
from pyzora.ring_secret import *
from pyzora.ring_types import *
from pyzora.child_behaviour_tools import *
__version__ = "1.0.0"
__author__ = "fortwoone"



"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

Game secret class. See GameSecret's docs for more details.

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>.
"""
import copy as mod_copy
from pyzora.secret import *


class GameSecret(BaseSecret):
    """A secret used to start a linked game.

    CAUTION : Non-Hero linked secrets are generated by one game and target the opposite one, but Hero's Secrets (non-linked)
    are generated by *the same game they are intended for*."""

    __link_name = "\0" * 5
    __child_name = mod_copy.deepcopy(__link_name)
    __behaviour = bytearray((0,))
    __animal = mod_copy.deepcopy(__behaviour)
    __target_game = mod_copy.deepcopy(__behaviour)
    __is_hero_quest = False
    __is_linked_game = False
    __was_given_free_ring = False

    __args = (
        "game_id",
        "region",
        "target_game",
        "link_name",
        "child_name",
        "animal",
        "behaviour",
        "is_linked_game",
        "is_hero_quest",
        "was_given_free_ring",
        "secret_array",
        "checksum",
        "secret_string",
    )

    def __init__(self, *args, **kwargs):
        for arg in self.__args:
            kwargs.setdefault(arg, None)
        for pos, arg in enumerate(self.__args):
            val = kwargs[arg]
            if val is None:
                try:
                    setattr(self, arg, args[pos])
                except LookupError:
                    pass
            else:
                try:
                    setattr(self, arg, val)
                except LookupError:
                    pass

    def __set_target_game(self, value: TargetGame | int):
        self.__target_game = TargetGame(value)

    target_game = property(lambda self: self.__target_game, __set_target_game,
                           doc="""The game the secret is meant for.
                           
                           If you attempt inputting a secret meant for Ages in Seasons,
                           for example, the input will fail.
                           
                           :type: TargetGame""")

    def __set_hero_quest(self, value: bool):
        self.__is_hero_quest = value

    is_hero_quest = property(lambda self: self.__is_hero_quest, __set_hero_quest,
                             doc="""A Hero's Secret allows the player to start
                             the game anew and get a special ring.
                             
                             :type: bool""")

    def __set_linked_game(self, value: bool):
        self.__is_linked_game = value

    is_linked_game = property(lambda self: self.__is_linked_game, __set_linked_game,
                              doc="""If a game is linked, some additional events will happen in the
                              target game.
                              
                              :type: bool""")

    def __set_link_name(self, value: str):
        if len(value.strip()) > 5:
            raise ValueError(f"incorrect name for Link : {value}")
        self.__link_name = value.strip().ljust(5, "\0").replace(" ", "\0")

    link_name = property(lambda self: self.__link_name.replace("\0", " "), __set_link_name,
                         doc="""Link's name. Takes at most 5 characters to fit in the secret.
                         
                         If you set this with a value containing spaces, they will be converted
                         into null characters for compatibility with the secrets.
                         
                         :type: str""")

    def __set_child_name(self, value: str):
        self.__child_name = value.strip().ljust(5, "\0")

    child_name = property(lambda self: self.__child_name.replace("\0", " "), __set_child_name,
                          doc="""Game secrets also store Bipin and Blossom's child's
                          name.
                          
                          If you set this with a value containing spaces, they will be converted
                          into null characters for compatibility with the secrets.
                          
                          :type: str""")

    def __set_animal(self, value: ObtainedCompanion | int):
        self.__animal = ObtainedCompanion(value)

    animal = property(lambda self: self.__animal, __set_animal,
                      doc="""The companion Link has obtained.
                      During linked games, this should be set.
                      In linked games, the companion stored in the secret will
                      recognise Link and give him his flute again.
                      
                      CAUTION: do NOT try to generate linked secrets with ObtainedCompanion.NONE as the animal
                      friend! While Ages does have a check to prevent you from getting stuck due to the lack of an animal 
                      friend and gives you Moosh if you input a secret with no animal friend, Seasons doesn't, and you will
                      get stuck.
                      
                      :type: ObtainedCompanion""")

    def __get_behaviour(self):
        return int(self.__behaviour[0])

    def __set_behaviour(self, value):
        self.__behaviour[0] = value

    behaviour = property(__get_behaviour, __set_behaviour,
                         doc="""Bipin and Blossom's child's behaviour is carried over
                         through secrets. 
                         
                         Note that the value stored here is NOT indicative of the child's 
                         behaviour on its own.
                         
                         :type: int""")

    def __set_free_ring_given(self, value: bool):
        self.__was_given_free_ring = value

    was_given_free_ring = property(lambda self: self.__was_given_free_ring, __set_free_ring_given,
                                   doc="""This should be set if the player has already received
                                   the Friendship Ring from Vasu.
                                   
                                   :type: bool""")

    @classmethod
    def load(cls, secret: bytes | bytearray | str, region: GameRegion) -> "GameSecret":
        """Load a game secret from bytes or a string.

        :param secret: The secret string/byte array to decode (if secret is an instance of str, then it will be parsed prior to loading).
        :type secret: str or bytearray
        :param region: The region to use when loading the secret.
        :type region: GameRegion
        :raise SecretError: if the given data is not 20 bytes long (only if used with bytes or bytearray, since strings will be parsed and reduced to 20 bytes).
        :raise NotAGameCodeError: if the given data is not for a game secret.
        :raise ChecksumError: if the given data's checksum doesn't match the expected one.
        :return: A secret based on the data contained in secret.
        :rtype: GameSecret
        """
        if isinstance(secret, str):
            # Secret string. Parse it before doing anything else.
            # Allows the user to give the region only once
            return cls.load(parse_secret(secret, region), region)
        if len(secret) != 20:
            raise SecretError("secret must contain exactly 20 bytes")
        decoded_bytes = cls.decode_bytes(secret, region)
        decoded_secret = byte_array_to_string(decoded_bytes)
        cloned_bytes = decoded_bytes.copy()
        cloned_bytes[19] = 0
        checksum = calculate_checksum(cloned_bytes)
        if decoded_bytes[19] & 0xF != checksum & 0xF:
            raise ChecksumError(
                f"checksum ({checksum}) does not match expected value({decoded_bytes[19]})"
            )
        del cloned_bytes
        game_id = int("".join(reversed(decoded_secret[5:20])), 2)
        if decoded_secret[3:5] != "00":
            raise NotAGameCodeError("given secret is not a game code")
        is_hero_quest, target_game = (func(itm) for func, itm in zip(("1".__eq__, int),
                                                                     decoded_secret[20:22]))
        is_linked_game = bool(int(decoded_secret[105]))
        link_name_array = bytearray((
            Byte(reverse_substring(decoded_secret, 22, 8)),
            Byte(reverse_substring(decoded_secret, 38, 8)),
            Byte(reverse_substring(decoded_secret, 60, 8)),
            Byte(reverse_substring(decoded_secret, 77, 8)),
            Byte(reverse_substring(decoded_secret, 89, 8))
        ))
        link_name = link_name_array.decode("utf-8")
        child_name_array = bytearray((
            Byte(reverse_substring(decoded_secret, 30, 8)),
            Byte(reverse_substring(decoded_secret, 46, 8)),
            Byte(reverse_substring(decoded_secret, 68, 8)),
            Byte(reverse_substring(decoded_secret, 97, 8)),
            Byte(reverse_substring(decoded_secret, 106, 8)),
        ))
        child_name = child_name_array.decode("utf-8")
        animal = ObtainedCompanion(Byte(reverse_substring(decoded_secret, 85, 4)))
        behaviour = Byte(reverse_substring(decoded_secret, 54, 6))
        was_given_free_ring = bool(int(decoded_secret[76]))
        return GameSecret(game_id=game_id, region=region,
                          link_name=link_name, target_game=target_game,
                          child_name=child_name, animal=animal,
                          was_given_free_ring=was_given_free_ring,
                          is_linked_game=is_linked_game,
                          is_hero_quest=is_hero_quest,
                          behaviour=behaviour)

    def __bytes__(self):
        link_byte_array = bytearray(self.link_name, "utf-8")
        child_byte_array = bytearray(self.child_name, "utf-8")
        cipher_key = (((self.game_id >> 8) + (self.game_id & 255)) & 7) * 2
        unencoded_secret = "".join(
            (
                reverse_string(integer_string(cipher_key).rjust(3, "0")),
                "00",  # This is to tell the game that we're writing a linking secret.
                reverse_string(integer_string(self.game_id).rjust(15, "0")),
                str(int(self.is_hero_quest)), str(int(self.target_game)),
                reverse_string(integer_string(link_byte_array[0]).rjust(8, "0")),
                reverse_string(integer_string(child_byte_array[0]).rjust(8, "0")),
                reverse_string(integer_string(link_byte_array[1]).rjust(8, "0")),
                reverse_string(integer_string(child_byte_array[1]).rjust(8, "0")),
                reverse_string(integer_string(self.behaviour).rjust(6, "0")),
                reverse_string(integer_string(link_byte_array[2]).rjust(8, "0")),
                reverse_string(integer_string(child_byte_array[2]).rjust(8, "0")),
                str(int(self.was_given_free_ring)),
                reverse_string(integer_string(link_byte_array[3]).rjust(8, "0")),
                reverse_string(integer_string(self.animal).rjust(4, "0")),
                reverse_string(integer_string(link_byte_array[4]).rjust(8, "0")),
                reverse_string(integer_string(child_byte_array[3]).rjust(8, "0")),
                str(int(self.is_linked_game)),
                reverse_string(integer_string(child_byte_array[4]).rjust(8, "0"))
            )
        )
        unencoded_bytes = string_to_byte_array(unencoded_secret)
        unencoded_bytes[19] = calculate_checksum(unencoded_bytes)
        secret = self._encode_bytes(unencoded_bytes, self.region)
        return bytes(secret)

    def __hash__(self):
        return hash(
            (
                self.game_id,
                self.__region,
                self.animal,
                self.behaviour,
                self.child_name,
                self.link_name,
                self.is_hero_quest,
                self.is_linked_game,
                self.target_game,
            )
        )



"""pyzora - Python library to help parsing secrets from Zelda OoS/OoA

Ring secret class. See RingSecret for more details.

(c) 2023 fortwoone.
All rights reserved.

This file is part of pyzora.

pyzora is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pyzora is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with pyzora. If not, see <https://www.gnu.org/licenses/>.
"""
from pyzora.secret import *
from pyzora.ring_types import *


class RingSecret(BaseSecret):
    """A ring secret. Ring secrets can be used to transfer a player's ring
    collection from one game to another.

    You can check if a ring is contained in a secret with the 'in' keyword.

    For example, if you want to check if the player already has the Friendship
    Ring, you can do this :

    if FRIENDSHIP in secret:
        do_whatever_you_want()"""
    __args = ("game_id", "rings", "ring_str", "region")
    __rings = 0
    __ring_str = ""

    def __set_rings(self, value: int):
        if value > int(AllRings):
            raise ValueError("cannot set a value higher than AllRings")
        if value < 0:
            raise ValueError("value must be an unsigned integer")
        # Converting the parameter into an integer in case we've been given a RingType instance
        self.__rings = int(value)
        self.__ring_str = bin(int(value))[2:].rjust(64, "0")

    rings = property(lambda self: self.__rings, __set_rings,
                     doc="""Get the rings stored in the secret as an integer.
                     
                     :type: int, AllRings or NoRings""")

    @property
    def ring_count(self):
        """Count how many rings the player has."""
        return self.__ring_str.count("1")

    def __init__(self, *args, **kwargs):
        for pos, arg in enumerate(self.__args):
            kwargs.setdefault(arg, None)
            if kwargs.get(arg) is None:
                try:
                    setattr(self, arg, args[pos])
                except LookupError:
                    pass
            else:
                try:
                    setattr(self, arg, kwargs.get(arg))
                except LookupError:
                    pass

    def __contains__(self, item):
        if item is NoRings:
            return not self.rings
        if item is AllRings:
            return self.rings == int(AllRings)
        return (item & self.rings) == item

    @classmethod
    def load(cls, secret: bytearray | bytes | str, region: GameRegion) -> "RingSecret":
        """Load a secret encoded with a certain region.

        :param secret: The secret string/byte array to decode (if secret is an instance of str, then it will be parsed prior to loading).
        :type secret: str or bytearray
        :param region: The region to use when loading the secret.
        :type region: GameRegion
        :raise SecretError: if the given data is not 15 bytes long (only if used with bytes or bytearray, since strings will be parsed and reduced to 15 bytes).
        :raise NotARingCodeError: if the given data is not for a ring secret.
        :raise ChecksumError: if the given data's checksum doesn't match the expected one.
        :return: A secret based on the data contained in secret.
        :rtype: RingSecret"""
        if isinstance(secret, str):
            # Secret string. Parse it before doing anything else.
            # Allows the user to give the region only once
            return cls.load(parse_secret(secret, region), region)
        if len(secret) != 15:
            raise SecretError("secret is expected to have exactly 15 bytes")
        decoded_bytes = cls.decode_bytes(secret, region)
        decoded_secret = byte_array_to_string(decoded_bytes)
        cloned_bytes = decoded_bytes.copy()
        cloned_bytes[14] = 0
        checksum = calculate_checksum(cloned_bytes)
        if (decoded_bytes[14] & 0xF) != (checksum & 0xF):
            raise ChecksumError(f"checksum {checksum} does not match expected value : {decoded_bytes[14]}")
        if decoded_secret[3] != "0" or decoded_secret[4] != "1":
            raise NotARingCodeError("given secret is not a ring code")
        game_id = int(reverse_substring(decoded_secret, 5, 15))
        ring_str = "".join((reverse_substring(decoded_secret, 36, 8),
                            reverse_substring(decoded_secret, 76, 8),
                            reverse_substring(decoded_secret, 28, 8),
                            reverse_substring(decoded_secret, 60, 8),
                            reverse_substring(decoded_secret, 44, 8),
                            reverse_substring(decoded_secret, 68, 8),
                            reverse_substring(decoded_secret, 20, 8),
                            reverse_substring(decoded_secret, 52, 8)))
        rings = int(ring_str, base=2)
        return RingSecret(game_id=game_id, rings=rings, ring_str=ring_str,
                          region=region)

    def __bytes__(self):
        ring_row1 = self.rings
        ring_row2 = ring_row1 >> 8
        ring_row3 = ring_row1 >> 16
        ring_row4 = ring_row1 >> 24
        ring_row5 = ring_row1 >> 32
        ring_row6 = ring_row1 >> 40
        ring_row7 = ring_row1 >> 48
        ring_row8 = ring_row1 >> 56
        cipher_key = ((self.game_id >> 8) + (self.game_id & 255)) & 7
        unencoded_secret = (
                reverse_string(integer_string(cipher_key).rjust(3, "0")) + "01"
        )
        secret_strings = map(reverse_string, (
            integer_string(self.game_id).rjust(15, "0"),
            integer_string(ring_row2).rjust(8, "0"),
            integer_string(ring_row6).rjust(8, "0"),
            integer_string(ring_row8).rjust(8, "0"),
            integer_string(ring_row4).rjust(8, "0"),
            integer_string(ring_row1).rjust(8, "0"),
            integer_string(ring_row5).rjust(8, "0"),
            integer_string(ring_row3).rjust(8, "0"),
            integer_string(ring_row7).rjust(8, "0"),
        ))
        unencoded_secret += "".join(secret_strings)
        unencoded_bytes = string_to_byte_array(unencoded_secret)
        del (
            ring_row1,
            ring_row2,
            ring_row3,
            ring_row4,
            ring_row5,
            ring_row6,
            ring_row7,
            ring_row8,
            secret_strings,
            unencoded_secret,
        )
        unencoded_bytes[14] = calculate_checksum(unencoded_bytes)
        secret = self._encode_bytes(unencoded_bytes, self.region)
        return bytes(secret)

    def to_list(self):
        """Return self as a list of ring types."""
        return list(filter(lambda tp: tp in self, RING_TYPES))

    def __hash__(self):
        return hash((self.__game_id, self.__rings))
